"""
municipality
============
"""

__all__ = ("Municipality",)

import dataclasses
from typing import List

import requests
import xmltodict
from cachetools import cached

from ..utils import BakalibError, _setup_logger, cache


class Municipality:
    """This is a container for certain methods for fetching cities and schools using Bakaláři school system
    
    :raises RuntimeError: If the class gets instantiated
    """

    municipality_url = "https://sluzby.bakalari.cz/api/v1/municipality/"
    logger = _setup_logger("[MUNICIPALITY]")

    @dataclasses.dataclass(frozen=True)
    class City:
        name: str
        school_count: int

    @dataclasses.dataclass(frozen=True)
    class School:
        id: str
        name: str
        url: str

    def __new__(cls, *args, **kwargs):
        raise RuntimeError(f"{cls} should not be instantiated!")

    @classmethod
    @cached(cache)
    def cities(cls) -> List["Municipality.City"]:
        """Fetches all cities in municipality database
        
        :raises BakalibError: If fails to retrieve cities
        :return: List of cities
        :rtype: List[Municipality.City]
        """
        try:
            resp = requests.get(cls.municipality_url, stream=True)
            parsed = xmltodict.parse(resp.content)
            return [
                cls.City(city["name"], int(city["schoolCount"]))
                for city in parsed["ArrayOfmunicipalityInfo"]["municipalityInfo"]
                if city["name"]
            ]
        except Exception as e:
            cls.logger.error(f"{type(e)}: {e}")
            raise BakalibError("Failed to retrieve city list from municipality")

    @classmethod
    def schools(cls, city: str) -> List["Municipality.School"]:
        """Fetches all schools in a specified city
        
        :param city: Name of a city
        :type city: str
        :raises BakalibError: If city not found
        :raises BakalibError: If failed to retrieve schools
        :return: List of schools
        :rtype: List[Municipality.School]
        """
        try:
            city_url = f"{cls.municipality_url}{city}"
            resp = requests.get(requests.utils.requote_uri(city_url), stream=True)

            if resp.status_code == 404:
                raise BakalibError("City not found")

            parsed = xmltodict.parse(resp.content)
            schools = parsed["municipality"]["schools"]["schoolInfo"]
            schools = schools if isinstance(schools, list) else [schools]

            return [
                cls.School(school["id"], school["name"], school["schoolUrl"])
                for school in schools
                if school["name"]
            ]
        except Exception as e:
            cls.logger.error(f"{type(e)}: {e}")
            raise BakalibError("Failed to retrieve school list from municipality")



"""
timetable
=========
"""

__all__ = ("TimetableModule",)

import datetime
import html
from dataclasses import dataclass
from typing import List
from cachetools import cached
from ..core.client import Client
from ..utils import _setup_logger, cache
from ._generic import GenericModule


class TimetableModule(GenericModule):
    """This is a class for accessing timetables of a client
    
    :param client: Instance of a client, defaults to None
    :type client: Client, optional
    :param url: URL of the school server, defaults to None
    :type url: str, optional
    :param token: Access token, defaults to None
    :type token: str, optional
    :param date: Initial date, defaults to datetime.date.today()
    :type date: datetime.date, optional
    """

    date: datetime.date

    @dataclass(frozen=True)
    class Header:
        caption: str
        time_begin: str
        time_end: str

    @dataclass(frozen=True)
    class Lesson:
        id_code: str
        type: str
        holiday: str
        abbr: str
        name: str
        teacher_abbr: str
        teacher: str
        room_abbr: str
        room: str
        absence_abbr: str
        absence: str
        theme: str
        group_abbr: str
        group: str
        cycle: str
        disengaged: str
        change_description: str
        notice: str
        caption: str
        time_begin: str
        time_end: str

    @dataclass(frozen=True)
    class Day:
        abbr: str
        date: str
        lessons: List["TimetableModule.Lesson"]

        def __len__(self):
            return len(self.lessons)

    @dataclass(frozen=True)
    class Result:
        headers: List["TimetableModule.Header"]
        days: List["TimetableModule.Day"]
        cycle_name: str

    def __init__(
        self,
        client: Client = None,
        url: str = None,
        token: str = None,
        date: datetime.date = datetime.date.today(),
    ):
        """Constructor method
        """
        super().__init__(client=client, url=url, token=token)
        self.date = date
        self.logger = _setup_logger(f"[TIMETABLE_{client.username}]")

    # ----------------------------------------------------

    def prev_week(self, prune: bool = True) -> "TimetableModule.Result":
        """Convenience method
        
        :param prune: Remove unnecessary empty lessons, defaults to True
        :type prune: bool, optional
        :return: Dataclass containing all days
        :rtype: TimetableModule.Result
        """
        self.date = self.date - datetime.timedelta(7)
        return self.date_week(self.date, prune)

    def this_week(self, prune: bool = True) -> "TimetableModule.Result":
        """Convenience method
        
        :param prune: Remove unnecessary empty lessons, defaults to True
        :type prune: bool, optional
        :return: Dataclass containing all days
        :rtype: TimetableModule.Result
        """
        self.date = datetime.date.today()
        return self.date_week(self.date, prune)

    def next_week(self, prune: bool = True) -> "TimetableModule.Result":
        """Convenience method
        
        :param prune: Remove unnecessary empty lessons, defaults to True
        :type prune: bool, optional
        :return: Dataclass containing all days
        :rtype: TimetableModule.Result
        """
        self.date = self.date + datetime.timedelta(7)
        return self.date_week(self.date, prune)

    # ----------------------------------------------------

    def date_week(
        self, date: datetime.date = None, prune: bool = True
    ) -> "TimetableModule.Result":
        """Fetches timetable data
        
        :param date: Date used to fetch the data, defaults to None
        :type date: datetime.date, optional
        :param prune: Remove unnecessary empty lessons, defaults to True
        :type prune: bool, optional
        :return: Dataclass containing all days
        :rtype: TimetableModule.Result
        """
        date_str = "{:04}{:02}{:02}".format(
            self.date.year, self.date.month, self.date.day
        )

        response = self.request(hx=self.token, pm="rozvrh", pmd=date_str)

        def holiday_check(obj: dict):
            zkratka = obj.get("zkratka")
            nazev = obj.get("nazev")
            if not zkratka:
                return nazev
            elif not nazev:
                return zkratka
            elif zkratka and nazev:
                return max([obj.get("zkratka"), obj.get("nazev")], key=len)
            else:
                return None

        headers = [
            self.Header(header["caption"], header["begintime"], header["endtime"])
            for header in response["rozvrh"]["hodiny"]["hod"]
        ]
        days = [
            self.Day(
                day["zkratka"],
                day["datum"],
                [
                    self.Lesson(
                        lesson.get("idcode"),
                        lesson.get("typ"),
                        holiday_check(lesson),
                        lesson.get("zkrpr"),
                        lesson.get("pr"),
                        lesson.get("zkruc"),
                        lesson.get("uc"),
                        lesson.get("zkrmist"),
                        lesson.get("mist"),
                        lesson.get("zkrabs"),
                        lesson.get("abs"),
                        html.unescape(lesson.get("tema"))
                        if lesson.get("tema")
                        else lesson.get("tema"),
                        lesson.get("zkrskup"),
                        lesson.get("skup"),
                        lesson.get("cycle"),
                        lesson.get("uvol"),
                        lesson.get("chng"),
                        lesson.get("notice"),
                        header.caption,
                        header.time_begin,
                        header.time_end,
                    )
                    for header, lesson in zip(headers, day["hodiny"]["hod"])
                ],
            )
            for day in response["rozvrh"]["dny"]["den"]
        ]

        if prune:
            lengths = []
            placeholder_lesson = None
            for day in days:
                for lesson in day.lessons:
                    if (
                        (lesson.type == "X" or lesson.type == "A")
                        and not lesson.holiday
                        and not lesson.change_description
                    ):
                        day.lessons.pop(0)
                    else:
                        break
                for lesson in reversed(day.lessons):
                    if (
                        (lesson.type == "X" or lesson.type == "A")
                        and not lesson.holiday
                        and not lesson.change_description
                    ):
                        day.lessons.pop()
                        placeholder_lesson = lesson
                    else:
                        break
                lengths.append(len(day))

            longest_day = None
            for day in sorted(days, key=len, reverse=True):
                if day.lessons[-1].type != "A" and day.lessons[-1].type != "X":
                    longest_day = day
                    break
            else:
                longest_day = max(days, key=len)

            headers = [
                self.Header(lesson.caption, lesson.time_begin, lesson.time_end)
                for lesson in (
                    longest_day.lessons if longest_day else max(days, key=len)
                )
            ]

            for day in days:
                while len(day.lessons) < max(lengths):
                    day.lessons.append(placeholder_lesson)

        return self.Result(headers, days, response["rozvrh"]["nazevcyklu"])

    def clear_cache(self) -> None:
        """Clears all entries related to timetable from cache
        """
        for entry in cache:
            if "rozvrh" in entry:
                cache.pop(entry)



"""
generic
=======
"""

__all__ = ("GenericModule",)

from ..core.client import Client
from ..utils import Base, BakalibError


class GenericModule(Base):
    """Generic module class, used by other modules.
        
        :param client: Instance of Client, defaults to None
        :type client: Client, optional
        :param url: URL of the school server, defaults to None
        :type url: str, optional
        :param token: Token of the client, defaults to None
        :type token: str, optional
        :raises BakalibError: If Client not logged in
        :raises BakalibError: If invalid module arguments
    """

    url: str
    token: str

    def __init__(
        self, client: Client = None, url: str = None, token: str = None
    ) -> None:
        """Constructor method
        """
        if client:
            if not client.logged_in:
                raise BakalibError("Client is not logged in.")
            self.url = client.url
            self.token = client.token
        elif url and token:
            self.url = url
            self.token = token
        else:
            raise BakalibError("Invalid module arguments.")



"""
util
====
"""

__all__ = ("BakalibError", "Base", "cache", "_setup_logger")

import datetime as _datetime
import inspect as _inspect
import logging as _logging
import pathlib as _pathlib
import sys as _sys

import cachetools as _cachetools
import requests as _requests
import xmltodict as _xmltodict

_log_dir = _pathlib.Path(_pathlib.Path.home() / ".bakalib")
_log_file = _pathlib.Path(
    _log_dir / (_datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S") + ".log")
)


if not _log_dir.is_dir():
    _log_dir.mkdir()
else:
    _logs = [f for f in _log_dir.iterdir()]
    _logs.sort()
    while len(_logs) > 5:
        _logs[0].unlink()
        _logs.pop(0)

_log_file.touch()


def _setup_logger(name: str) -> _logging.Logger:
    """Sets up a logger with the provided name
    
    :param name: Name of the logger
    :type name: str
    :return: Ready-to-use logger
    :rtype: _logging.Logger
    """
    formatter = _logging.Formatter(
        fmt="%(asctime)s — %(name)s — %(levelname)s — %(message)s — %(funcName)s:%(lineno)d"
    )
    handler = _logging.FileHandler(_log_file, mode="a")
    handler.setFormatter(formatter)
    logger = _logging.getLogger(name)
    logger.setLevel(_logging.DEBUG)
    logger.addHandler(handler)
    return logger


_logger = _setup_logger("default")

cache = _cachetools.LFUCache(32)


class BakalibError(Exception):
    """:class:`Exception` subclass, used for differentiating between Python exceptions and bakalib exceptions
    """

    pass


class Base:
    """Base class for most of the classes present in this library
    """

    @_cachetools.cached(cache)
    def request(self, **kwargs: str) -> dict:
        """Generic request method
        
        :param gethx: Name of the client
        :type gethx: str, optional
        :param hx: Access token
        :type hx: str, optional
        :param pm: Module name
        :type pm: str, optional
        :param pmd: Module arguments
        :type pmd: str, optional
        :raises BakalibError: If received response is invalid
        :return: Received response
        :rtype: dict
        """
        resp = _requests.get(url=self.url, params=kwargs, verify=False)
        results: dict = _xmltodict.parse(resp.content).get("results")

        try:
            res: dict = results.get("res")
            result: dict = results.get("result")
        except AttributeError as e:
            log_args: dict = kwargs.copy()
            log_args.pop(k="hx", d=None)
            log_args.pop(k="gethx", d=None)
            _logger.error(log_args)

        comp: str = res if res else result

        if not comp == "01":
            raise BakalibError("Received response is invalid")

        return results



"""
grades
======
"""

__all__ = ("GradesModule",)

from dataclasses import dataclass
from typing import List

from ..core.client import Client
from ..utils import BakalibError, _setup_logger
from ._generic import GenericModule


class GradesModule(GenericModule):
    """This is a class for accessing grades of a client
    
    :param client: Instance of Client, defaults to None
    :type client: Client, optional
    :param url: URL of the school server, defaults to None
    :type url: str, optional
    :param token: Access token of a client, defaults to None
    :type token: str, optional
    """

    @dataclass(frozen=True)
    class Grade:
        subject: str
        max_points: str
        grade: str
        gr: str
        points: str
        date: str
        date_granted: str
        weight: str
        caption: str
        note: str
        type: str
        description: str

    @dataclass(frozen=True)
    class Subject:
        name: str
        abbr: str
        average_round: str
        average: str
        recalculation: str
        points_to_grade: str
        quarter: str
        note: str
        glob_note: str
        grades: List["GradesModule.Grade"]

        def __len__(self):
            return len(self.grades)

    def __init__(self, client: Client = None, url: str = None, token: str = None):
        """Constructor method
        """
        super().__init__(client=client, url=url, token=token)
        self.logger = _setup_logger(f"[GRADES_{client.username}]")

    def subjects(self) -> List["GradesModule.Subject"]:
        """Fetches all subjects and grades
        
        :raises BakalibError: If no grades returned
        :return: Dataclass containing subjects list
        :rtype: GradesModule.Result
        """
        response = self.request(hx=self.token, pm="znamky")
        if response["predmety"] is None:
            raise BakalibError("Grades module returned None, no grades were found.")

        for index, subject in enumerate(response["predmety"]["predmet"]):
            if not isinstance(subject["znamky"]["znamka"], list):
                response["predmety"]["predmet"][index]["znamky"]["znamka"] = [
                    subject["znamky"]["znamka"]
                ]

        return [
            self.Subject(
                subject["nazev"],
                subject["zkratka"],
                subject["prumer"],
                subject["numprumer"],
                subject["prepocet"],
                subject["bodytoznm"],
                subject["ctvrt"],
                subject["pozn"],
                subject["globpozn"],
                [
                    self.Grade(
                        grade.get("pred"),
                        grade.get("maxb"),
                        grade.get("znamka"),
                        grade.get("zn"),
                        grade.get("bd"),
                        grade.get("datum"),
                        grade.get("udeleno"),
                        grade.get("vaha"),
                        grade.get("caption"),
                        grade.get("poznamka"),
                        grade.get("typ"),
                        grade.get("ozn"),
                    )
                    for grade in subject["znamky"]["znamka"]
                ],
            )
            for subject in response["predmety"]["predmet"]
        ]



"""
client module

contains:

    Client: class

"""

__all__ = ("Client",)

import base64
import datetime
import hashlib
from dataclasses import dataclass

from ..utils import Base, BakalibError, _setup_logger


def _is_logged_in(invert: bool = False):
    """Checks if arg[0] (Client) is logged in.
    
    :param invert: Inverts the check, defaults to False
    :type invert: bool, optional
    :raises BakalibError: If Client is not logged in
    :raises BakalibError: If Client is already logged in
    """

    def decorator(func):
        def wrapper(*args, **kwargs):
            if args[0].logged_in ^ invert:
                return func(*args, **kwargs)
            else:
                if invert:
                    raise BakalibError("Client is already logged in.")
                raise BakalibError("Client is not logged in.")

        return wrapper

    return decorator


class Client(Base):
    """This is a class representing a user on a school server. It needs to be logged in (:meth:`.Client.login`) before using with other components. 
        
    :param username: Name of the user
    :type username: str
    :param url: URL address of the school server
    :type url: str
    """

    username: str
    url: str

    perm_token: str
    token: str

    logged_in: bool

    @dataclass(frozen=True)
    class Info:
        version: str
        name: str
        type_abbr: str
        type: str
        school: str
        school_type: str
        class_: str
        year: str
        modules: str
        newmarkdays: str

    def __init__(self, username: str, url: str):
        """Constructor method
        """
        self.username = username
        self.url = url

        self.perm_token = None
        self.token = None

        self.logged_in = False

        self.logger = _setup_logger(f"[CLIENT_{self.username}]")
        self.logger.debug(f"INITIALIZED")

    def __str__(self):
        return f"Client(username={self.username}, url={self.url})"

    def __repr__(self):
        return {"username": self.username, "url": self.url}

    @_is_logged_in(invert=True)
    def login(
        self, password: str = None, perm_token: str = None, check_valid: bool = True
    ):
        """Logs the client in
        
        :param password: Password, defaults to None
        :type password: str, optional
        :param perm_token: Permanent token generated from password, defaults to None
        :type perm_token: str, optional
        :param check_valid: Checks if the session is valid after logging in, defaults to True
        :type check_valid: bool, optional
        :raises BakalibError: If incorrect arguments are provided
        :raises BakalibError: If generated login token is invalid (wrong password/perm_token)
        """
        if perm_token:
            self.perm_token = perm_token
            token = self._token(self.perm_token)
        elif password:
            self.perm_token = self._permanent_token(password)
            token = self._token(self.perm_token)
        else:
            raise BakalibError("Incorrect arguments")

        if check_valid:
            if self._is_token_valid(token):
                self.token = token
            else:
                raise BakalibError("Token is invalid: Invalid password/perm_token")
        else:
            self.token = token

        self.logged_in = True
        self.logger.info(f"LOGGED IN SUCCESSFULLY")

    def _permanent_token(self, password: str) -> str:
        """Generates a permanent token which can be used for login
        
        :param password: Password
        :type password: str
        :raises BakalibError: If the username is not found on the server
        :return: Permanent token
        :rtype: str
        """
        self.logger.debug("GENERATING PERM_TOKEN")
        try:
            resp = self.request(gethx=self.username)
        except BakalibError as e:
            self.logger.warning("PERM_TOKEN IS INVALID")
            raise BakalibError("Invalid username")

        salt = resp.get("salt")
        ikod = resp.get("ikod")
        typ = resp.get("typ")
        salted_password = f"{salt}{ikod}{typ}{password}".encode("utf-8")
        hashed_password = base64.b64encode(hashlib.sha512(salted_password).digest())
        perm_token = (
            f"*login*{self.username}*pwd*{hashed_password.decode('utf8')}*sgn*ANDR"
        )
        self.logger.debug("PERM_TOKEN GENERATED")
        return perm_token

    def _token(self, perm_token: str) -> str:
        """Generates an access token using current time
        
        :param perm_token: Permanent token generated by :meth:`.Client._permanent_token`
        :type perm_token: str
        :return: Access token
        :rtype: str
        """
        self.logger.debug("GENERATING ACCESS TOKEN")
        today = datetime.date.today()
        datecode = "{:04}{:02}{:02}".format(today.year, today.month, today.day)
        token_hash = hashlib.sha512((perm_token + datecode).encode("utf-8")).digest()
        token = base64.urlsafe_b64encode(token_hash).decode("utf-8")
        self.logger.debug("ACCESS TOKEN GENERATED")
        return token

    def _is_token_valid(self, token: str) -> bool:
        """Checks if generated token from :meth:`.Client._token` is valid
        
        :param token: Access token
        :type token: str
        :return: True of False, depends on the result
        :rtype: bool
        """
        try:
            self.request(hx=token, pm="login")
            self.logger.info(f"TOKEN IS VALID")
            return True
        except BakalibError:
            self.logger.warning(f"TOKEN IS INVALID")
            return False

    @_is_logged_in()
    def info(self) -> "Client.Info":
        """Fetches information about the current user
        
        :return: Dataclass containing user information
        :rtype: Client.Info
        """

        response = self.request(hx=self.token, pm="login")
        result = self.Info(
            *[
                response.get(element).get("newmarkdays")
                if element == "params"
                else response.get(element)
                for element in response
                if not element == "result"
            ]
        )
        return result
